import os
import time
from threading import Thread
import logging

import nvidia_smi as nvml

from grn.core.globals import GPUStates


log = logging.getLogger(__file__)


__all__ = ['GPUMonitor', 'SingleGPUProcessMonitor']


class GPUMonitor(Thread):
    """
    """
    def __init__(self, delay=1):
        super().__init__()

        nvml.nvmlInit()
        try:
            self.num_gpus_available = nvml.nvmlDeviceGetCount()
        except nvml.NVMLError:
            raise LookupError(
                f'Could not acquire device count, meaning there are no NVIDIA devices available.')

        # The GPU master must respect external CUDA_VISIBLE_DEVICES restrictions.
        self.cuda_visible_devices = os.environ.get('CUDA_VISIBLE_DEVICES')
        if self.cuda_visible_devices is not None:
            self.cuda_visible_devices = set(int(d) for d in self.cuda_visible_devices.split(','))
        else:
            self.cuda_visible_devices = set(range(self.num_gpus_available))

        log.info(f'Managing GPUs {self.cuda_visible_devices}\n')
 
        try:
            self.gpu_handles = {
                idx: nvml.nvmlDeviceGetHandleByIndex(idx) 
                for idx in range(self.num_gpus_available)
                if idx in self.cuda_visible_devices}
        except nvml.NVMLError:
            raise LookupError(
                f'Could not acquire one or more GPU device handles.')

        self.stopped = False
        self.delay = delay
        self.start()

    def _update_gpu_states(self, states):
        # The lock is probably necessary here because of the sequence of appends.
        # We don't want to get interrupted by the server trying to heappop a gpu.
        with GPUStates.LOCK:
            gpu_list = []
            for gid in states:
                GPUStates.STATES[gid] = states[gid]
                gpu_list.append( (gid, states[gid][GPUStates.TARGET_RESOURCE]) )

            # Heapify at the end to for O(n) instead of O(n log n).
            GPUStates.GPU_QUEUE.clear()
            GPUStates.GPU_QUEUE.extend(gpu_list)

        GPUStates.STATES_INIT.set()

    def run(self):
        while not self.stopped:
            try:
                states = {
                    idx: (
                        nvml.nvmlDeviceGetMemoryInfo(h), 
                        nvml.nvmlDeviceGetUtilizationRates(h)) 
                    for idx, h in self.gpu_handles.items()}
            except nvml.NVMLError:
                break

            states = {
                idx: dict(
                    memoryUsed=info[0].used,
                    memoryTotal=info[0].total,
                    memoryFree=info[0].free,
                    load=info[1].gpu) 
                for idx, info in states.items()}
            self._update_gpu_states(states)

            time.sleep(self.delay)

    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        nvml.nvmlShutdown()
        self.stopped = True


class SingleGPUProcessMonitor(Thread):
    """Separate thread for monitoring usage stats of a target GPU.
    """
    def __init__(self, gpu_id: int, pid: int, delay: float):
        super().__init__()
        nvml.nvmlInit()
        try:
            self.gpu_handle = nvml.nvmlDeviceGetHandleByIndex(gpu_id)
        except nvml.NVMLError:
            raise LookupError(
                f'Could not acquire device handle for GPU with index {gpu_id}.')

        self.stopped = False
        self.delay = delay
        self.gpu_id = gpu_id
        self.pid = pid
        self.max_mem_used = 0
        # self.mem_total = 0
        # self.max_load = 0
        self.start()

    def run(self):
        while not self.stopped:
            used_memory = None
            try:
                # info = nvml.nvmlDeviceGetMemoryInfo(self.gpu_handle)
                # res = nvml.nvmlDeviceGetUtilizationRates(self.gpu_handle)
                procs = nvml.nvmlDeviceGetComputeRunningProcesses(self.gpu_handle)
                for proc in procs:
                    if proc.pid == self.pid:
                        used_memory = proc.usedGpuMemory
                        break
            except nvml.NVMLError:
                break

            if used_memory is not None:
                # Aggregate peak usage
                self.max_mem_used = max(self.max_mem_used, used_memory)

            time.sleep(self.delay)
    
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        nvml.nvmlShutdown()
        self.stopped = True



import os
import sys
import signal
import functools
import inspect
import contextlib
import logging
import time

import grpc

from grn.utils.utils import is_server_available, find_gpu_master_address
from grn.utils.gpu_monitors import SingleGPUProcessMonitor
import grn.grpc_resources.master_pb2 as protos
import grn.grpc_resources.master_pb2_grpc as services
from grn.core.constants import ServiceErrorCode, ResourcePolicy

from typing import (
    Callable, 
    Iterable, 
    Tuple, 
    Any,
    Optional,
)


log = logging.getLogger(__file__)


__all__ = ['job']


def __grpc_failed_job_handler(
    *args, 
    stub: services.GPUMasterStub, 
    jobtype: protos.JobType
) -> None:
    profile = protos.JobProfile(
        jobtype=jobtype,
        succeeded=False)
    stub.CompleteJob(profile, wait_for_ready=True)
    sys.exit()


@contextlib.contextmanager
def __grpc_handle_signals(
    sigs: Iterable[signal.Signals],
    stub: services.GPUMasterStub, 
    jobtype: protos.JobType, 
) -> None:
    orig_handlers = [signal.getsignal(s) for s in sigs]
    for s in sigs:
        signal.signal(s, functools.partial(
            __grpc_failed_job_handler, 
            stub=stub, jobtype=jobtype))
    yield

    # Restore control to original handlers
    for s, handler in zip(sigs, orig_handlers):
        signal.signal(s, handler)


def __request_gpu(
    stub: services.GPUMasterStub, 
    jobtype: protos.JobType
) -> int:
    response: protos.GPU = stub.RequestGPU(jobtype, wait_for_ready=True)
    errorcode = ServiceErrorCode(response.errorcode)

    # Dumb retry policy where we just keep checking over and over again
    # for available resources. The increasing delay period helps reduce a
    # bit of waste, but still isn't great.
    # TODO: Would be much better to do some kind of event-driven approach where
    #       jobs can efficiently wait until resources free up rather than wasting
    #       cycles.
    delay = 0.1
    max_delay = 5.0
    while ( (errorcode == ServiceErrorCode.EXCEEDS_CURRENT_MEMORY) or 
            (errorcode == ServiceErrorCode.WAITING_FOR_JOB_PROFILE) ):

        time.sleep(min(max_delay, delay))
        delay *= 2

        response: protos.GPU = stub.RequestGPU(jobtype, wait_for_ready=True)
        errorcode = ServiceErrorCode(response.errorcode)

    if errorcode == ServiceErrorCode.EXCEEDS_TOTAL_MEMORY:
        raise MemoryError(f'{errorcode}: Cannot complete job \n```\n{jobtype}```\n')
    return response.gpu_id


def __run_job(
    job_func: Callable, 
    gpu_id: int, 
    jobtype: protos.JobType
) -> Tuple[Any, protos.JobProfile]:
    os.environ['CUDA_VISIBLE_DEVICES'] = f'{gpu_id}'

    # Collects job profile stats in separate thread.
    with SingleGPUProcessMonitor(gpu_id, pid=os.getpid(), delay=0.01) as monitor: 
        mem_used = None
        try:
            outputs = job_func()
        except RuntimeError as e:
            outputs = None
            profile = protos.JobProfile(
                jobtype=jobtype,
                succeeded=False)
        else:
            # if mem_used is None:
            #     mem_used = monitor.max_mem_used
            # else:
            #     mem_used = max(monitor.max_mem_used, mem_used)
            mem_used = monitor.max_mem_used
            if mem_used is None:
                raise SystemError(
                    f'No usage metrics could be collected for PID {os.getpid()} '
                    f'on GPU {gpu_id} because no such PID was ever found by NVML.'
                    f'Make sure that you are monitoring the device ID on which this '
                    f'process was launched.'
                    f'\nNOTE: If you are running in a docker container, make sure to'
                    f'use the --pid=host flag to turn off PID namespace isolation. '
                    f'NVML uses the host PID namespace regardless of the container settings.')

            profile = protos.JobProfile(
                jobtype=jobtype,
                succeeded=True,
                gpu=protos.GPU(
                    gpu_id=gpu_id, 
                    errorcode=0),
                max_gpu_memory_used=mem_used)

    return outputs, profile


def job(
    jobstr: Optional[str] = None,
    resource_policy: str = 'spread',
    continue_on_server_unavailable: bool = False,
) -> Callable:
    if resource_policy in {'spread', 'spreading'}:
        __RESOURCE_POLICY = ResourcePolicy.SPREAD.value
    elif resource_policy in {'pack', 'packing'}:
        __RESOURCE_POLICY = ResourcePolicy.PACK.value
    else:
        raise ValueError(f'Got resource_policy={resource_policy}, but options are (\'spread\', \'pack\').')

    def load_balance(
        func: Callable,
    ) -> Callable:
        # Construct the jobstr that specifies the job type.
        func_file = inspect.getfile(func)
        __JOBSTR = jobstr or (
            f'{func_file}::{func.__class__.__name__}' if inspect.isclass(func)
            else f'{func_file}::{func.__name__}')
        __JOBTYPE = protos.JobType(
            jobstr=__JOBSTR, 
            resource_policy=__RESOURCE_POLICY)

        __REGISTRY = {}
        # This function is added as an attribute of the decorator at the end.
        # This allows it to be invoked as a decorator itself.
        def __profile(pfunc: Callable) -> Callable:
            if __REGISTRY.get('profiler') is not None:
                raise AttributeError(
                    f'Tried to register profiler {inspect.getfile(pfunc)}::{pfunc.__name__} '
                    f'for {__JOBSTR}, but one already exists: '
                    f'{inspect.getfile(__REGISTRY["profiler"])}::{__REGISTRY["profiler"].__name__}')
            __REGISTRY['profiler'] = pfunc
            return pfunc
 
        @functools.wraps(func)
        def decorator(*args, **kwargs):
            try:
                __GPU_MASTER_ADDR = find_gpu_master_address()
            except FileNotFoundError as e:
                if not continue_on_server_unavailable:
                    raise e
                __GPU_MASTER_ADDR = 'None'

            # If the server is unavailable, then raise exception by default.
            # Otherwise, proceed with computation without load balancing.
            if not is_server_available(__GPU_MASTER_ADDR):
                if not continue_on_server_unavailable:
                    raise ConnectionError('GPU Master is not available.')
                log.warn(
                    'GPU Master is not running but continue_on_server_unavailable=True, '
                    'proceeding anyway... This may result in unexpected job failures.')
                return func(*args, **kwargs)

            with grpc.insecure_channel(
                __GPU_MASTER_ADDR, 
                options=(('grpc.enable_http_proxy', 0),)
            ) as channel:
                stub = services.GPUMasterStub(channel)

                # We need to communicate abnormal process termination to the server.
                # TODO: Using a gRPC stream should eliminate the need for this signal
                #       handling.
                with __grpc_handle_signals(
                    (signal.SIGINT, signal.SIGTERM), stub, __JOBTYPE
                ):
                    # Run the profile job if this is the first invocation of this job
                    # relative to the server lifetime.
                    # If no profiler was registered, then run the full job as the profile
                    # job and return its outputs.
                    is_job_profiled = stub.JobTypeExists(__JOBTYPE).response
                    if not is_job_profiled: 
                        job_func = __REGISTRY.get(
                            (lambda: __REGISTRY['profiler'](*args, **kwargs)), 
                            (lambda: func(*args, **kwargs)))
                        gpu_id = __request_gpu(stub, __JOBTYPE)
                        outputs, profile = __run_job(
                            job_func=job_func,
                            gpu_id=gpu_id, 
                            jobtype=__JOBTYPE)
                        stub.CompleteJob(profile)

                    # If a profiler was registered, now run the actual job and 
                    # return its outputs. If the profile failed, don't run the full
                    # job.
                    # The other case is that this job type has already been profiled,
                    # which means we should just run it.
                    if (
                        is_job_profiled or (
                            (not is_job_profiled) and 
                            (profile.succeeded) and 
                            (__REGISTRY.get('profiler') is not None))
                    ):
                        gpu_id = __request_gpu(stub, __JOBTYPE)
                        outputs, profile = __run_job(
                            job_func=(lambda: func(*args, **kwargs)),
                            gpu_id=gpu_id,
                            jobtype=__JOBTYPE)
                        stub.CompleteJob(profile)

            return outputs

        decorator.profile = __profile
        return decorator
    return load_balance



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import master_pb2 as master__pb2


class GPUMasterStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RequestGPU = channel.unary_unary(
                '/GPUMaster/RequestGPU',
                request_serializer=master__pb2.JobType.SerializeToString,
                response_deserializer=master__pb2.GPU.FromString,
                )
        self.CompleteJob = channel.unary_unary(
                '/GPUMaster/CompleteJob',
                request_serializer=master__pb2.JobProfile.SerializeToString,
                response_deserializer=master__pb2.Empty.FromString,
                )
        self.JobTypeExists = channel.unary_unary(
                '/GPUMaster/JobTypeExists',
                request_serializer=master__pb2.JobType.SerializeToString,
                response_deserializer=master__pb2.BoolResponse.FromString,
                )


class GPUMasterServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RequestGPU(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CompleteJob(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def JobTypeExists(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GPUMasterServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RequestGPU': grpc.unary_unary_rpc_method_handler(
                    servicer.RequestGPU,
                    request_deserializer=master__pb2.JobType.FromString,
                    response_serializer=master__pb2.GPU.SerializeToString,
            ),
            'CompleteJob': grpc.unary_unary_rpc_method_handler(
                    servicer.CompleteJob,
                    request_deserializer=master__pb2.JobProfile.FromString,
                    response_serializer=master__pb2.Empty.SerializeToString,
            ),
            'JobTypeExists': grpc.unary_unary_rpc_method_handler(
                    servicer.JobTypeExists,
                    request_deserializer=master__pb2.JobType.FromString,
                    response_serializer=master__pb2.BoolResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'GPUMaster', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class GPUMaster(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RequestGPU(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GPUMaster/RequestGPU',
            master__pb2.JobType.SerializeToString,
            master__pb2.GPU.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CompleteJob(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GPUMaster/CompleteJob',
            master__pb2.JobProfile.SerializeToString,
            master__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def JobTypeExists(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GPUMaster/JobTypeExists',
            master__pb2.JobType.SerializeToString,
            master__pb2.BoolResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import depq

from threading import Lock, Event
from typing import (
    Dict, 
    Union, 
    Optional,
)
NumericType = Union[int, float]


__all__ = ['GPUStates', 'JobStates']


class GPUStates:
    STATES: Dict[int, Dict[str, NumericType]] = dict()
    STATES_INIT = Event()
    LOCK = Lock()
    GPU_QUEUE: depq.DEPQ = depq.DEPQ()
    PROFILING_GROUP: Dict[int, str] = dict()

    # TODO: This should probably be defined in constants.py
    TARGET_RESOURCE: str = 'memoryUsed'


class JobStates:
    JOB_TYPES = dict()
    ACTIVE_JOBS = dict()
    MAX_NUM_JOBS: Optional[int] = None
    LOCK = Lock()


import pathlib
import tempfile
from enum import Enum, auto

__all__ = ['GRN_SERVER_TMP_INFO_PATH', 'ServiceErrorCode']


# Obtain a universally consistent location to cache intermediate info needed by clients.
GRN_SERVER_TMP_INFO_PATH = pathlib.Path(tempfile.gettempdir()) / '.grn.conf'

class ResourcePolicy(Enum):
    SPREAD = auto()
    PACK = auto()

class ServiceErrorCode(Enum):
    OK = auto()
    EXCEEDS_TOTAL_MEMORY = auto()
    EXCEEDS_CURRENT_MEMORY = auto()
    WAITING_FOR_JOB_PROFILE = auto()